import numpy as np
import pyfftw
import multiprocessing


class Field(object):

    def __init__(self, colors, ndim=6):

        # Dimension of the basic field quantity - 6 for symmetric 3x3 matrices such as strains and stresses, 3 for vectors, such as temp gradient and heat flux
        self.ndim = ndim

        # Dimensions of the microstructure
        Nx, Ny, Nz = np.shape(colors)
        self.shape = (Nx,Ny,Nz)
        
        # Initialization of the real and complex fields
        # The FFTW (our chosen FFT package) is able to work 'inplace'
        # Hence we put the (slightly smaller) real field in the same memory as the complex field
        # Effectively the same memory can be read/written as real or complex numbers

        # The arrays are 4-dimensional. The first dimension gives you the component of the field quantity (e.g., the strain)
        # The remaining quantities are associated with the coordinates of the voxel (real field) or the frequency (complex field)
        # Example: real_field[:,2,3,4] gives you the 6 components at (2,3,4)

        self.cpx_field = pyfftw.zeros_aligned((self.ndim, Nx, Ny, Nz // 2 + 1),
                                              dtype=np.complex128)
        
        self.real_field = np.ndarray(shape=(self.ndim, Nx, Ny, Nz),
                                dtype=np.float64,
                                buffer=self.cpx_field.data)

        # Number of used threads
        threadCount = multiprocessing.cpu_count()

        # FFTW creates an optimal 'plan' for the FFT
        self.fftPlans = pyfftw.FFTW(self.real_field, self.cpx_field,
                                    axes=(1, 2, 3),
                                    threads=threadCount)

        # Same for the inverse FFT
        self.ifftPlans = pyfftw.FFTW(self.cpx_field, self.real_field,
                                     direction='FFTW_BACKWARD',
                                     axes=(1, 2, 3),
                                     threads=threadCount)
        self.status = 'real'

    def initialize(self,strain):
        for i in range(self.ndim):
            self.real_field[i] = strain[i]
            
    # Methods to apply the FFT and IFFT
    # In Fourier space the cpx_field should be accessed (e.g., for applying the Gamma^0 operator)
    # In real space the real_field should be accessed -> see self.status for current status
    
    def fft(self):
        self.fftPlans()
        self.status = 'fourier'

    def ifft(self):

        self.ifftPlans()
        self.status = 'real'